/*
   Program: PythonLauncher v1.00.
   Purpose: Locates and launches a Python interpreter for a Python script.

   Copyright (c) 2000, Toolwood.  All Rights Reserved.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name Toolwood not be used in
   advertising or publicity pertaining to distribution of the software without
   specific, written prior permission by Toolwood.

   TOOLWOOD DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
   ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
   TOOLWOOD BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
   ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
   WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
   ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
   SOFTWARE.

   Please also be advised that this source file is not representative of
   commercial Toolwood source code.  Toolwood developer's use a set of very
   advanced libraries developed by Toolwood.  These libraries include features
   such as advanced exception handling (even in C), which is not used by this
   source code, which again means the error handling is nowhere near as good
   as that of commercial Toolwood code.  In other words: You got this stuff
   for free, so don't complain about the somewhat insufficient error handling.
   This also means that this source file includes non-portable features, which
   would be normally be isolated in the Toolwood Runtime Library.

   Revision history:
      2000.08.05, Mikael Lyngvig (mikael.lyngvig@toolwood.dk):
         Eliminated dependencies on Toolwood in-house libraries so that
         the source code can be made available to the public.
      2000.07.09, Mikael Lyngvig (mikael.lyngvig@toolwood.dk):
         Initial version.
*/

#include <windows.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>


#define SEP  "\\"
#define ARRAYSIZE(Array)  (sizeof(Array) / sizeof((Array)[0]))


#define ERROR_PYLAUNCH_PATH_INVALID         1001
#define ERROR_PYLAUNCH_SCRIPT_UNKNOWN       1002
#define ERROR_PYLAUNCH_INTERPRETER_UNKNOWN  1003
#define ERROR_PYLAUNCH_INTERPRETER_EXECUTE  1004


/* puts:
   Alternate to STDIO's puts() function.  This variant (hopefully) eliminates
   the need to link in the entire IO subsystem and thus reduces the size of
   the final executable.
*/
void puts(const char *pValue)
{
   DWORD Written;

   WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), pValue, strlen(pValue), &Written, NULL);
}


/* execute:
   Invokes the command `pCommand` and returns the integer result in `pResult`.
   The result code is application dependent, but usually zero means "okay" and
   everything else means "an error occured".
*/
int execute(const char *pCommand, int *pResult)
{
   HANDLE hProcess = NULL;
   STARTUPINFO StartupInfo;
   PROCESS_INFORMATION ProcessInfo;
   int rc;

   GetStartupInfo(&StartupInfo);
   rc = CreateProcess(
      NULL,                            /* lpApplicationName */
      (LPTSTR) pCommand,               /* lpCommandLine */
      NULL,                            /* lpProcessAttributes */
      NULL,                            /* lpThreadAttributes */
      TRUE,                            /* bInheritanceFlag */
      0,                               /* dwCreationFlags */
      NULL,                            /* lpEnvironment */
      NULL,                            /* lpCurrentDirectory */
      &StartupInfo,                    /* lpStartupInfo */
      &ProcessInfo                     /* lpProcessInformation */
   );
   if (rc == 0)
      return -1;

   WaitForSingleObject(ProcessInfo.hProcess, INFINITE);

   GetExitCodeProcess(ProcessInfo.hProcess, pResult);

   CloseHandle(ProcessInfo.hThread);
   CloseHandle(ProcessInfo.hProcess);

   return 0;
}


/* file_exists:
   Returns TRUE if the specified file exists and is a disk file.
*/
int file_exists(const char *pName)
{
   HANDLE hFile = NULL;
   int Found = 0;

   hFile = CreateFile(pName, 0, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
   if (hFile == NULL)
      goto failed;

   Found = (GetFileType(hFile) == FILE_TYPE_DISK);
   CloseHandle(hFile);

failed:
   return Found;
}


/* path_getdir:
   Extracts the path portion (drive + directory) and returns it in `pResult`.
*/
void path_getdir(const char *pValue, size_t MaxLen, char *pResult)
{
   char Dir[_MAX_DIR + 1];

   assert(MaxLen >= 2);
   _splitpath(pValue, pResult, Dir, NULL, NULL);
   assert(MaxLen >= strlen(pResult) + strlen(Dir));

   strcat(pResult, Dir);
}


/* path_getname:
   Extracts the name portion (no extension) and returns it in `pResult`.
*/
void path_getname(const char *pValue, size_t MaxLen, char *pResult)
{
   _splitpath(pValue, NULL, NULL, pResult, NULL);
   assert(strlen(pResult) <= MaxLen);
}


/* path_getext:
   Extracts the (last) extension and returns it in `pResult`.
*/
void path_getext(const char *pValue, size_t MaxLen, char *pResult)
{
   _splitpath(pValue, NULL, NULL, NULL, pResult);
   assert(strlen(pResult) <= MaxLen);
}


/* path_absolute:
   Expands the specified path into its fully qualified equivalent and returns
   the result in `pResult`.
   If unable to expand the path, the error code ERROR_PYLAUNCH_PATH_INVALID
   is returned.
*/
int path_absolute(const char *pValue, size_t MaxLen, char *pResult)
{
   int rc = 0;

   if (_fullpath(pResult, pValue, MaxLen) == NULL)
      rc = ERROR_PYLAUNCH_PATH_INVALID;

   return rc;
}


/* string_quote:
   Quotes the specified string, if necessary, and returns the result in
   `pResult`.
   The string is quoted if a character not in the string `Unquotable` is
   found.
*/
void string_quote(const char *pValue, char *pResult)
{
   int Quote = FALSE;
   const char *pString;
   static const char Unquotable[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890:.,!@#$^*-+?~_;\\\'`";

   assert(pValue != NULL);
   assert(pResult != NULL);
   assert(pResult != pValue);

   for (pString = pValue; *pString != '\0'; pString += 1)
   {
      if (strchr(Unquotable, *pString) == NULL)
      {
         Quote = TRUE;
         break;
      }
   }

   if (Quote)
      *pResult++ = '\"';
   strcpy(pResult, pValue);
   if (Quote)
      strcat(pResult, "\"");
}


/* string_split:
   Splits the specified string and appends the result to the `apResult` array.
   The array index variable, `pnItems`, is incremented equivalently.
*/
void string_split(const char *pValue, char Sep, char *apResult[], unsigned *pnItems)
{
   char *pListCopy = NULL;
   char *pSource;
   const char *pSubString;           /* ptr to start of current substring */
   unsigned nItems;
   char ch;

   assert(pValue != NULL);
   assert(Sep != '\0');
   assert(apResult != NULL);
   assert(pnItems != NULL);

   nItems = *pnItems;

   pListCopy = strdup(pValue);
   pSource   = pListCopy;

   pSubString = NULL;

   do
   {
      ch = *pSource;

      if (ch == Sep || ch == '\0')
      {
         if (pSubString != NULL && pSource > pSubString)
         {
            *pSource = '\0';
            apResult[nItems++] = strdup(pSubString);
            pSubString = NULL;
         }
      }
      else if (pSubString == NULL)
         pSubString = pSource;

      pSource += 1;
   } while (ch != 0);

   free(pListCopy);
   *pnItems = nItems;
}


int main(int argc, const char *argv[])
{
   int rc = 0;
   char ExeName[_MAX_PATH + 1];
   char PathName[_MAX_PATH + 1];
   char BaseName[_MAX_PATH + 1];
   const char *pPathStr = getenv("PATH");
   int Found;
   char *apPaths[128];
   unsigned    nPaths = 0;
   char ScriptName[_MAX_PATH + 1];
   char InterpreterName[_MAX_PATH + 1];
   char TmpStr[_MAX_PATH + 1];

   assert(argc >= 1);
   assert(argv != NULL);

   /* split executable path into useful components */
   rc = path_absolute(argv[0], _MAX_PATH, ExeName);
   if (rc != 0)
      goto failed;
   path_getdir(ExeName, _MAX_PATH, PathName);
   path_getname(ExeName, _MAX_PATH, BaseName);

   /* create list of directories to search */
   if (getenv("PYTHONHOME") != NULL)
      apPaths[nPaths++] = strdup(getenv("PYTHONHOME"));
   apPaths[nPaths++] = strdup(PathName);
   string_split(pPathStr, ';', apPaths, &nPaths);
   assert(nPaths < ARRAYSIZE(apPaths));

   /* find the Python script */
   {
      static const char *apExtensions[] = { "py", "pyc", "pyo" };
      unsigned iPath;

      for (iPath = 0; iPath < nPaths; iPath += 1)
      {
         char Name[_MAX_PATH + 1];
         unsigned iExt;

         for (iExt = 0; iExt < ARRAYSIZE(apExtensions); iExt += 1)
         {
            strcpy(Name, apPaths[iPath]);
            if (Name[strlen(Name) - 1] != SEP[0])
               strcat(Name, SEP);
            strcat(Name, BaseName);
            strcat(Name, ".");
            strcat(Name, apExtensions[iExt]);

            if (file_exists(Name))
            {
               strcpy(ScriptName, Name);
               goto ScriptFound;
            }
         }
      }

      /* if we get here, we didn't find the script */
      rc = ERROR_PYLAUNCH_SCRIPT_UNKNOWN;
      goto failed;

ScriptFound:
      (void) NULL;
   }

   /* find the Python interpreter */
   {
      static const char *apExecutables[] = { "cgipython.exe", "python.exe", "pythonw.exe" };
      unsigned iPath;

      for (iPath = 0; iPath < nPaths; iPath += 1)
      {
         char Name[_MAX_PATH + 1];
         unsigned iExe;

         for (iExe= 0; iExe < ARRAYSIZE(apExecutables); iExe += 1)
         {
            strcpy(Name, apPaths[iPath]);
            if (Name[strlen(Name) - 1] != SEP[0])
               strcat(Name, SEP);
            strcat(Name, apExecutables[iExe]);

            if (file_exists(Name))
            {
               strcpy(InterpreterName, Name);
               goto InterpreterFound;
            }
         }
      }

      /* if we get here, we didn't find the script */
      rc = ERROR_PYLAUNCH_INTERPRETER_UNKNOWN;
      goto failed;

InterpreterFound:
      (void) NULL;
   }

   {
      int wrc;
      char Command[2048];
      int iArg;

      strcpy(Command, InterpreterName);
      strcat(Command, " ");
      strcat(Command, ScriptName);
      for (iArg = 1; iArg < argc; iArg += 1)
      {
         strcat(Command, " ");
         string_quote(argv[iArg], TmpStr);
         strcat(Command, TmpStr);
      }

      wrc = execute(Command, &rc);
      if (wrc == -1)
      {
         rc = ERROR_PYLAUNCH_INTERPRETER_EXECUTE;
         goto failed;
      }

      /* return interpreter's exit code */
      return rc;
   }

failed:
   {
      const char *pMessage = NULL;

      switch (rc)
      {
         case ERROR_PYLAUNCH_PATH_INVALID :
            pMessage = "Error: Invalid path";
            break;

         case ERROR_PYLAUNCH_SCRIPT_UNKNOWN :
            pMessage = "Error: Unable to locate Python script";
            break;

         case ERROR_PYLAUNCH_INTERPRETER_UNKNOWN :
            pMessage = "Error: Unable to locate Python interpreter";
            break;

         case ERROR_PYLAUNCH_INTERPRETER_EXECUTE :
         case -1 :
            pMessage = "Error: Unable to execute Python interpreter";
            break;

         default :
            pMessage = "Internal Error: Unhandled failure detected";
            break;
      }

      puts(pMessage);

   }

   /* return error code */
   return 1;
}
